@@ -7,32 +7,14 @@ import zipfile
 import pandas as pd
 import requests
 
-from sa_gwdata.identifiers import parse_well_ids, UnitNo, ObsNo, Well, Wells
-
-__all__ = (
-    "WaterConnectSession",
-    "Response",
-    "find_wells",
-    "find_wells_in_lat_lon",
-    "wells_summary",
-    "water_levels",
-    "salinities",
-    "water_chem",
-    "construction_events",
-    "construction_details",
-    "drillers_logs",
-    "strat_logs",
-    "hydrostrat_logs",
-    "lith_logs",
-    "elevation_surveys",
-)
+from sa_gwdata.identifiers import *
 
 
 logger = logging.getLogger(__name__)
 
 
 class Response(object):
-    def __init__(self, response, **kwargs):
+    def __init__(self, response, working_crs="EPSG:4326", **kwargs):
         """Groundwater Data HTTP response.
 
         Args:
@@ -40,6 +22,7 @@ class Response(object):
 
         """
         self.response = response
+        self.working_crs = working_crs
         for key, value in kwargs.items():
             setattr(self, key, value)
 
@@ -60,15 +43,30 @@ class Response(object):
         """If the response is a list, convert to a pandas DataFrame with
         columns converted into lowercase."""
         if not hasattr(self, "_df"):
-            df = pd.DataFrame(self.json).rename(columns=str.lower)
-            self._df = df
+            if len(self.json) > 0:
+                if "DHNO" in self.json[0]:
+                    if isinstance(self.json[0]["DHNO"], list):
+                        # e.g. GetSuburbFromName
+                        convert_object = self.json[0]["DHNO"]
+                    else:
+                        convert_object = self.json
+                else:
+                    # e.g. GetPWASearchData
+                    convert_object = self.json
+
+                df = pd.DataFrame(convert_object).rename(columns=str.lower)
+                self._df = df
+                logger.debug(f"Response has length={len(df)}")
+            else:
+                logger.debug("Response is empty")
         return self._df
 
     @property
     def df_exists(self):
         """Check if JSON can be converted to a DataFrame. Returns bool."""
         if isinstance(self.json, list):
-            return True
+            if len(self.json) > 0:
+                return True
         return False
 
     def gdf(self, x_col="lon", y_col="lat"):
@@ -77,11 +75,11 @@ class Response(object):
 
         df = self.df
         return gpd.GeoDataFrame(
-            df, geometry=df.apply(lambda x: Point(x[x_col], x[y_col]), axis=1)
-        )
+            df, geometry=gpd.points_from_xy(df[x_col], df[y_col]), crs="EPSG:4326"
+        ).to_crs(self.working_crs)
 
 
-def get_global_session(force_new=False):
+def get_global_session(working_crs="EPSG:4326", force_new=False):
     if not "__waterconnect_session" in globals():
         global __waterconnect_session
         __waterconnect_session = WaterConnectSession()
@@ -92,450 +90,6 @@ def get_global_session(force_new=False):
     return __waterconnect_session
 
 
-def find_wells(input_text, **kwargs):
-    """Find wells and retrieve some summary information.
-
-    Args:
-        input_text (str): any well identifiers to parse. See
-            :func:`sa_gwdata.parse_well_ids_plaintext` for details of
-            other keyword arguments you can pass here.
-
-    For example:
-
-        >>> import sa_gwdata
-        >>> wells = sa_gwdata.find_wells("yat99 5840-46 ULE205")
-        ...
-        >>> wells
-        ['MLC008', 'ULE205', 'YAT099']
-
-    """
-    session = get_global_session()
-    return session.find_wells(input_text, **kwargs)
-
-
-def find_wells_in_lat_lon(lats, lons, **kwargs):
-    """Find wells within a geographic area.
-
-        Args:
-            lats (tuple): two decimal latitudes
-            lons (tuple): two decimal longitudes
-
-        Returns: a list of :class:`sa_gwdata.Well` objects (actually
-            a :class:`sa_gwdata.Wells` object which very closely
-            resembles a list).
-
-        """
-    session = get_global_session()
-    return session.find_wells_in_lat_lon(lats, lons, **kwargs)
-
-
-def wells_summary(wells, session=None, **kwargs):
-    """Get table of summary information for wells.
-
-    Args:
-        wells (list): list of drillhole numbers (ints)
-            or :class:`sa_gwdata.Well` objects
-
-    Returns: pandas DataFrame with these columns:
-
-        - dh_no (int64)
-        - unit_hyphen (str)
-        - unit_long (int64)
-        - dh_name (str)
-        - network (str)
-        - obs_no (str)
-        - drillhole_class (str)
-        - aquifer (str)
-        - orig_drilled_depth (float64)
-        - orig_drilled_date (datetime64[ns])
-        - max_drilled_depth (float64)
-        - max_drilled_date (datetime64[ns])
-        - latest_open_depth (float64)
-        - latest_open_date (datetime64[ns])
-        - latest_permit_no (float64)
-        - cased_to (float64)
-        - casing_min_diam (float64)
-        - purpose (str)
-        - latest_status (str)
-        - latest_status_date (datetime64[ns])
-        - latest_dtw (float64)
-        - latest_swl (float64)
-        - latest_rswl (float64)
-        - latest_wl_date (datetime64[ns])
-        - latest_tds (float64)
-        - latest_ec (float64)
-        - latest_sal_date (datetime64[ns])
-        - latest_ph (float64)
-        - latest_ph_date (datetime64[ns])
-        - latest_yield (float64)
-        - latest_yield_date (datetime64[ns])
-        - easting (float64)
-        - northing (float64)
-        - zone (int64)
-        - lon_deg (int64)
-        - lon_min (int64)
-        - lon_sec (float64)
-        - lat_deg (int64)
-        - lat_min (int64)
-        - lat_sec (float64)
-        - longitude (float64)
-        - latitude_positive (float64)
-        - latitude (float64)
-        - hundred (str)
-        - plan (str)
-        - parcel (str)
-        - cert_title (str)
-        - map_250k (str)
-        - map_100k (int64)
-        - map_50k (int64)
-        - map_10k (int64)
-        - map_2_5k (str)
-        - map_1k (int64)
-        - water_info_exists (str)
-        - salinity_exists (str)
-        - water_chem_exists (str)
-        - geophys_log_exists (str)
-        - drillers_log_exists (str)
-        - lith_log_exists (str)
-
-    """
-    session = get_global_session()
-    return session.bulk_wells_summary(wells, **kwargs)
-
-
-def water_levels(wells, session=None, **kwargs):
-    """Get table of water level measurements for wells.
-
-    Args:
-        wells (list): list of drillhole numbers (ints)
-            or :class:`sa_gwdata.Well` objects
-
-    Returns: pandas DataFrame with these columns:
-
-        - dh_no (int64)
-        - network (str)
-        - unit_long (int64)
-        - aquifer (str)
-        - easting (float64)
-        - northing (float64)
-        - zone (int64)
-        - unit_hyphen (str)
-        - obs_no (str)
-        - obs_date (datetime64[ns])
-        - dtw (float64)
-        - swl (float64)
-        - rswl (float64)
-        - pressure (float64)
-        - temperature (float64)
-        - dry_ind (str)
-        - anomalous_ind (str)
-        - pump_ind (str)
-        - measured_during (str)
-        - data_source (str)
-        - comments (str)
-
-    """
-    session = get_global_session()
-    return session.bulk_water_levels(wells, **kwargs)
-
-
-def salinities(wells, session=None, **kwargs):
-    """Get table of salinity samples for wells.
-
-    Args:
-        wells (list): list of drillhole numbers (ints)
-            or :class:`sa_gwdata.Well` objects
-
-    Returns: pandas DataFrame with these columns:
-
-        - dh_no (int64)
-        - network (str)
-        - aquifer (str)
-        - unit_hyphen (str)
-        - unit_long (int64)
-        - obs_no (str)
-        - collected_date (datetime64[ns])
-        - collected_time (str)
-        - tds (int64)
-        - ec (int64)
-        - ph (float64)
-        - sample_type (str)
-        - anomalous_ind (str)
-        - test_place (str)
-        - extract_method (str)
-        - measured_during (str)
-        - data_source (str)
-        - easting (float64)
-        - northing (float64)
-        - zone (int64)
-
-    """
-    if session is None:
-        session = get_global_session()
-    return session.bulk_salinities(wells, **kwargs)
-
-
-def water_chem(wells, session=None, **kwargs):
-    """Get table of water chemistry sample analyses for wells.
-
-    Args:
-        wells (list): list of drillhole numbers (ints)
-            or :class:`sa_gwdata.Well` objects
-
-    Returns: pandas DataFrame with these columns:
-
-        - dh_no (int64)
-        - unit_hyphen (str)
-        - obs_no (str)
-        - sample_no (int64)
-        - collected_date (datetime64[ns])
-        - analysis_code (str)
-        - analysis_name (str)
-        - value (float64)
-        - unit (str)
-
-    """
-    if session is None:
-        session = get_global_session()
-    return session.bulk_water_chem(wells, **kwargs)
-
-
-def elevation_surveys(wells, session=None, **kwargs):
-    """Get table of elevation surveys for wells.
-
-    Args:
-        wells (list): list of drillhole numbers (ints)
-            or :class:`sa_gwdata.Well` objects
-
-    Returns: pandas DataFrame with these columns:
-
-        - dh_no (int64)
-        - unit_hyphen (str)
-        - obs_no (str)
-        - network (str)
-        - elev_date (datetime64[ns])
-        - ground_elev (float64)
-        - ref_elev (float64)
-        - survey_meth (str)
-        - vert_accuracy (float64)
-        - ref_point_type (str)
-        - applied_date (datetime64[ns])
-        - comments (str)
-
-    """
-    if session is None:
-        session = get_global_session()
-    return session.bulk_elevation_surveys(wells, **kwargs)
-
-
-def construction_events(wells, session=None, **kwargs):
-    """Get table of construction events (summary data) for wells.
-
-    Args:
-        wells (list): list of drillhole numbers (ints)
-            or :class:`sa_gwdata.Well` objects
-
-    Returns: pandas DataFrame with these columns:
-
-        - dh_no (int64)
-        - unit_hyphen (str)
-        - obs_no (str)
-        - completion_date (datetime64[ns])
-        - total_depth (float64)
-        - final_depth (float64)
-        - current_depth (float64)
-        - permit_no (float64)
-        - backfilled (str)
-        - casing_from (float64)
-        - casing_to (float64)
-        - casing_min_diam (float64)
-        - casing_material (str)
-        - pcemented (str)
-        - pcement_from (float64)
-        - pcement_to (float64)
-        - pzone_from (float64)
-        - pzone_to (float64)
-        - pzone_min_diam (float64)
-        - pzone_type (str)
-        - pzone_material (str)
-        - pzone_aperture (float64)
-        - drill_from (float64)
-        - drill_to (float64)
-        - drill_diam (float64)
-        - drill_method (str)
-        - development_method (str)
-        - development_duration (float64)
-        - comments (str)
-
-    """
-    if session is None:
-        session = get_global_session()
-    return session.bulk_construction_events(wells, **kwargs)
-
-
-def construction_details(wells, session=None, **kwargs):
-    """Get table of detailed information collected at time of construction of wells.
-
-    Args:
-        wells (list): list of drillhole numbers (ints)
-            or :class:`sa_gwdata.Well` objects
-
-    Returns: dict with four pandas DataFrames. The keys, and the
-        columns of the DataFrames are below:
-
-        * "water_cuts":
-            - dh_no (int64)
-            - unit_hyphen (str)
-            - water_cut_date (datetime64[ns])
-            - depth_from (float64)
-            - depth_to (float64)
-            - swl (float64)
-            - yield (float64)
-            - test_method (str)
-            - tds (float64)
-            - ec (float64)
-            - sample_type (str)
-
-        * "drilling":
-            - dh_no (int64)
-            - unit_hyphen (str)
-            - depth_from (float64)
-            - depth_to (float64)
-            - hole_diam (float64)
-            - drill_method (str)
-            - comments (str)
-
-        * "casing":
-            - dh_no (int64)
-            - unit_hyphen (str)
-            - depth_from (float64)
-            - depth_to (float64)
-            - casing_diam (int64)
-            - casing_material (str)
-            - cement_method (str)
-            - cement_from (float64)
-            - cement_to (float64)
-            - comments (str)
-
-        * "prod_zones":
-            - dh_no (int64)
-            - unit_hyphen (str)
-            - pzone_type (str)
-            - depth_from (float64)
-            - depth_to (float64)
-            - pzone_diam (float64)
-            - pzone_material (str)
-            - pzone_aperture (float64)
-            - comments (str)
-
-
-    """
-    if session is None:
-        session = get_global_session()
-    return session.bulk_construction_details(wells, **kwargs)
-
-
-def drillers_logs(wells, session=None, **kwargs):
-    """Get table of lithological intervals from drillers logs for wells.
-
-    Args:
-        wells (list): list of drillhole numbers (ints)
-            or :class:`sa_gwdata.Well` objects
-
-    Returns: pandas DataFrame with these columns:
-
-        - dh_no (int64)
-        - unit_hyphen (str)
-        - obs_no (str)
-        - log_date (datetime64[ns])
-        - logger_name (str)
-        - depth_from (float64)
-        - depth_to (float64)
-        - lith_code (str)
-        - description (str)
-
-    """
-    if session is None:
-        session = get_global_session()
-    return session.bulk_drillers_logs(wells, **kwargs)
-
-
-def strat_logs(wells, session=None, **kwargs):
-    """Get table of formations from the stratigraphic logs for wells.
-
-    Args:
-        wells (list): list of drillhole numbers (ints)
-            or :class:`sa_gwdata.Well` objects
-
-    Returns: pandas DataFrame with these columns:
-
-        - dh_no (int64)
-        - unit_hyphen (str)
-        - depth_from (float64)
-        - depth_to (float64)
-        - strat_name (str)
-        - gis_code (str)
-
-    """
-    if session is None:
-        session = get_global_session()
-    return session.bulk_strat_logs(wells, **kwargs)
-
-
-def hydrostrat_logs(wells, session=None, **kwargs):
-    """Get table of formations from the hydrostratigraphic logs for wells.
-
-    Args:
-        wells (list): list of drillhole numbers (ints)
-            or :class:`sa_gwdata.Well` objects
-
-    Returns: pandas DataFrame with these columns:
-
-        - dh_no (int64)
-        - unit_hyphen (str)
-        - unit_depth_from (float64)
-        - unit_depth_to (float64)
-        - subunit_depth_from (float64)
-        - subunit_depth_to (float64)
-        - hydro_type (float64)
-        - hydro_depth_to_greater_flag (str)
-        - comments (str)
-        - map_symbol (str)
-        - strat_name (str)
-        - subunit_comments (float64)
-        - subunit_code (float64)
-        - subunit_desc (float64)
-
-    """
-    if session is None:
-        session = get_global_session()
-    return session.bulk_hydrostrat_logs(wells, **kwargs)
-
-
-def lith_logs(wells, session=None, **kwargs):
-    """Get table of lithological logs for wells.
-
-    Args:
-        wells (list): list of drillhole numbers (ints)
-            or :class:`sa_gwdata.Well` objects
-
-    Returns: pandas DataFrame with these columns:
-
-        - dh_no (int64)
-        - unit_hyphen (str)
-        - obs_no (str)
-        - depth_from (float64)
-        - depth_to (float64)
-        - major_lith_code (str)
-        - minor_lith_code (str)
-        - descr (str)
-
-    """
-    if session is None:
-        session = get_global_session()
-    return session.bulk_lith_logs(wells, **kwargs)
-
-
 class WaterConnectSession:
     """Wrapper around repeated requests to Groundwater Data.
 
@@ -562,9 +116,17 @@ class WaterConnectSession:
     }
 
     def __init__(
-        self, *args, endpoint=None, sleep=2, verify=True, load_list_data=True, **kwargs
+        self,
+        *args,
+        endpoint=None,
+        sleep=2,
+        verify=True,
+        load_list_data=True,
+        working_crs="EPSG:4326",
+        **kwargs,
     ):
-        self.well_cache = pd.DataFrame(columns=set(self.well_id_cols.values()))
+        self.well_cache = pd.DataFrame(columns=list(set(self.well_id_cols.values())))
+        self.working_crs = working_crs
         self.verify = verify
         if not endpoint:
             endpoint = "https://www.waterconnect.sa.gov.au/_layouts/15/dfw.sharepoint.wdd/{app}.ashx/"
@@ -595,11 +157,19 @@ class WaterConnectSession:
             path = self.endpoint + path
         path = path.format(app=app)
         logger.debug("GET {} verify={}".format(path, verify))
-        response = requests.get(path, verify=verify, **kwargs)
+        try:
+            response = requests.get(path, verify=verify, **kwargs)
+        except requests.exceptions.ConnectionError:
+            time.sleep(0.8)
+            response = requests.get(path, verify=verify, **kwargs)
         self.last_request = time.time()
         endpoint, name = path.rsplit("/", 1)
         logger.debug("Response content = {}".format(response.content))
-        return self._cache_data(Response(response, endpoint=endpoint, name=name))
+        return self._cache_data(
+            Response(
+                response, working_crs=self.working_crs, endpoint=endpoint, name=name
+            )
+        )
 
     def post(self, path, app="WDDDMS", verify=None, **kwargs):
         if verify is None:
@@ -615,7 +185,9 @@ class WaterConnectSession:
         response = requests.post(path, verify=verify, **kwargs)
         self.last_request = time.time()
         endpoint, name = path.rsplit("/", 1)
-        return Response(response, endpoint=endpoint, name=name)
+        return Response(
+            response, working_crs=self.working_crs, endpoint=endpoint, name=name
+        )
 
     def bulk_download_wells(self, service, wells, **kwargs):
         return bulk_download(service, {"DHNOs": wells.dh_no}, **kwargs)
@@ -696,12 +268,15 @@ class WaterConnectSession:
         )
         return df
 
-    def bulk_water_levels(self, wells, **kwargs):
+    def bulk_water_levels(self, wells, pumping=True, anomalous=True, **kwargs):
         dh_nos = [w for w in wells]
         if len(wells):
             if hasattr(wells[0], "dh_no"):
                 dh_nos = [w.dh_no for w in wells]
-        df = self.bulk_download("GetWaterLevelDownload", {"DHNOs": dh_nos})
+        df = self.bulk_download(
+            "GetWaterLevelDownload",
+            {"DHNOs": dh_nos, "Pumping": pumping, "Anomalous": anomalous},
+        )
         df["obs_date"] = pd.to_datetime(df.obs_date, format="%d/%m/%Y")
         df = df.rename(
             columns={
@@ -845,7 +420,7 @@ class WaterConnectSession:
         resp = self.post(
             "GetConstructionDetailsDownload?bulkOutput=CSV",
             data={"exportdata": json.dumps({"DHNOs": [w.dh_no for w in wells]})},
-            **kwargs
+            **kwargs,
         )
         dfs = {}
         name_map = {
@@ -1032,8 +607,9 @@ class WaterConnectSession:
         r2 = self.get(
             "GetObswellNumberSearchData", params={"OBSNUMBER": ",".join(obs_nos)}
         )
+        dfs = [r.df for r in [r1, r2] if r.df_exists]
         df = (
-            pd.concat([r1.df, r2.df], sort=False)
+            pd.concat(dfs, sort=False)
             .drop_duplicates()
             .rename(
                 columns={"dhno": "dh_no", "mapnum": "unit_no", "obsnumber": "obs_no"}
@@ -1071,6 +647,7 @@ class WaterConnectSession:
         self.aquifers = {
             item["V"]: item["T"].replace((item["V"] + ": "), "")
             for item in response.json["Aquifer"]
+            if "V" in item
         }
         self.networks = {item["V"]: item["T"] for item in response.json["Networks"]}
         self.nrm_regions = {
@@ -1095,3 +672,33 @@ class WaterConnectSession:
         # SALINITY - can join by  AND
         # GetPWASearchData?PWA=Angas-Bremer&Q=SALSTATUS='C'
         # GetPWASearchData?PWA=Angas-Bremer&Q=SWLSTATUS='C'%20AND%20SALSTATUS='C'
+
+    def search_by_suburb(self, suburb):
+        suburb = suburb.upper()
+        return self.get("GetSuburbFromName", app="WDDWFS", params={"Suburb": suburb})
+
+    def search_by_radius(self, lat, lon, radius):
+        # GetCircleData?Box=-34.155353,138.455916&Radius=2
+        lat = float(lat)
+        lon = float(lon)
+        radius = float(radius)
+        return self.get("GetCircleData", params={"Box": [lat, lon], "Radius": radius})
+
+    def search_by_rect(self, sw_corner, ne_corner):
+        lat0, lon0 = sw_corner
+        lat1, lon1 = ne_corner
+        return self.get(
+            "GetGridData",
+            params="Box=" + ",".join([str(c) for c in [lat1, lon0, lat0, lon1]]),
+        )
+
+    def search_by_network(self, *network):
+        network_codes = [n for n in network if n in self.networks.keys()]
+        networks_inverted = {v: k for k, v in self.networks.items()}
+        network_codes += [
+            networks_inverted[n] for n in network if n in self.networks.values()
+        ]
+        logger.debug(f"Querying for validated network codes: {network_codes}")
+        return self.get(
+            "GetObswellNetworkData", params="Network=" + ",".join(network_codes)
+        )
